/*  $Id: button_boring.cpp,v 1.5 2003/09/18 17:53:29 mbn Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
**
*/

#include "API/Display/surface.h"
#include "API/Display/font.h"
#include "API/Core/Resources/resource_manager.h"
#include "API/Core/Math/rect.h"
#include "API/GUIStyleBoring/stylemanager_boring.h"
#include "API/GUIStyleBoring/button_boring.h"

// need for highlight checks
#include "API/GUI/gui_manager.h"

/////////////////////////////////////////////////////////////////////////////
// Construction:

CL_Button_Boring::CL_Button_Boring(
	CL_Button *button,
	CL_StyleManager_Boring *style)
:
	CL_ComponentStyle(button),
	button(button),
	style(style),
	font(0), 
	font_disabled(0),
        delete_font(false),
        delete_font_disabled(false),
	down(false)
{
	resources = style->get_resources();

	if(resources->exists("Button/font"))
		set_font(new CL_Font("Button/font", resources), true);
	if(resources->exists("Button/font_disabled"))
		set_font_disabled(new CL_Font("Button/font_disabled", resources), true);

	slot_set_options = button->sig_set_options().connect(
		this, &CL_Button_Boring::on_set_options);
	slot_paint = button->sig_paint().connect(
		this, &CL_Button_Boring::on_paint);
	slot_get_preferred_size = button->sig_get_preferred_size().connect(
		this, &CL_Button_Boring::on_get_preferred_size);
	slot_pressed = button->sig_pressed().connect(
		this, &CL_Button_Boring::on_pressed);
	slot_released = button->sig_released().connect(
		this, &CL_Button_Boring::on_released);
}

CL_Button_Boring::~CL_Button_Boring()
{
	if(delete_font && font)
		delete font;
	if(delete_font_disabled && font_disabled)
		delete font_disabled;
}

/////////////////////////////////////////////////////////////////////////////
// Operations:

void CL_Button_Boring::set_font(CL_Font *_font, bool _delete_font)
{
	if(delete_font && font)
		delete font;
	font = _font;
	delete_font = _delete_font;
}

void CL_Button_Boring::set_font_disabled(CL_Font *_font, bool _delete_font)
{
	if(delete_font_disabled && font_disabled)
		delete font_disabled;
	font_disabled = _font;
	delete_font_disabled = _delete_font;
}

/////////////////////////////////////////////////////////////////////////////
// Callbacks:

void CL_Button_Boring::on_set_options(const CL_DomElement &options)
{
	if (options.has_attribute("font"))
		set_font(new CL_Font(options.get_attribute("font"), resources), true);
	if (options.has_attribute("font_disabled"))
		set_font_disabled(new CL_Font(options.get_attribute("font_disabled"), resources), true);
}

void CL_Button_Boring::on_get_preferred_size(CL_Point &size)
{
	if(font)
	{
		size.x = font->get_width(button->get_text()) + 8;
		size.y = font->get_height(button->get_text()) + 4;
	}
}

void CL_Button_Boring::on_paint()
{
	int text_width = 0;
	int text_height = 0;

	if(font)
	{
		text_width = font->get_width(button->get_text().c_str());
		text_height = font->get_height();
	}

	int button_width = button->get_width();
	int button_height = button->get_height();

	int x_pos = (button_width - text_width) / 2;
	int y_pos = (button_height - text_height) / 2;

	int width = button->get_width();
	int height = button->get_height();

	if(button->is_enabled() == false)
	{
		// Fill
		style->fill_rect(2, 2, width - 2, height - 2, GUICOLOR_BUTTON_DISABLED);

		// Outline
		style->draw_rect(0, 0, width, height, GUICOLOR_DARK_OUTLINE_DISABLED);

		// Shade box
		style->draw_box(1, 1, width - 1 , height - 1, GUICOLOR_BRIGHT_SHADE_DISABLED, GUICOLOR_DARK_SHADE_DISABLED);

		if(font_disabled)
			font_disabled->draw(x_pos, y_pos, button->get_text());
		else if(font)
			font->draw(x_pos, y_pos, button->get_text());
	}
	else 
	{
		bool being_clicked = button->has_mouse_over() ? down : false;
		if((button->is_down() && !being_clicked) || (!button->is_down() && being_clicked))
		{
			// Fill
			style->fill_rect(2, 2, width - 2, height - 2, GUICOLOR_BUTTON_TOGGLED);

			// Outline
			style->draw_rect(0, 0, width, height, GUICOLOR_DARK_OUTLINE);

			// Shade box
			style->draw_box(1, 1, width - 1 , height - 1, GUICOLOR_DARK_SHADE, GUICOLOR_BRIGHT_SHADE);

			if(font)
				font->draw(x_pos + 1, y_pos + 1, button->get_text());
		}
		else
		{
			bool need_highlight = false;
			if (button->has_mouse_over())
				if (!button->get_gui_manager() ||
					(!button->get_gui_manager()->get_modal_component() ||
					button->has_modal_parent()))
					need_highlight = true;

			// Fill
			style->fill_rect(2, 2, width - 2, height - 2, GUICOLOR_BUTTON);

			// Outline
			style->draw_rect(0, 0, width, height, GUICOLOR_DARK_OUTLINE);

			// Shade box
			style->draw_box(1, 1, width - 1 , height - 1, GUICOLOR_BRIGHT_SHADE, GUICOLOR_DARK_SHADE);

			// Focus
			if(button->has_focus())
				style->draw_rect(2, 2, width - 2, height - 2, GUICOLOR_FOCUS);

			if(font)
				font->draw(x_pos, y_pos, button->get_text());
		}
	}
}

void CL_Button_Boring::on_pressed()
{
	down = true;
}

void CL_Button_Boring::on_released()
{
	down = false;
}
