/*  $Id: input_device_sdljoystick.cpp,v 1.5 2003/10/05 11:40:05 mbn Exp $
**
**  ClanLib Game SDK
**  Copyright (C) 2003  The ClanLib Team
**  For a total list of contributers see the file CREDITS.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
*/

#include <cstdio>

#include "API/Display/input_device.h"
#include "API/Display/input_event.h"
#include "input_device_sdljoystick.h"
#include "API/Core/System/error.h"
#include "API/Core/System/log.h"

#include "display_window_sdl.h"

#include <SDL/SDL.h>

/////////////////////////////////////////////////////////////////////////////
// CL_InputDevice_SDLJoystick construction:

CL_InputDevice_SDLJoystick::CL_InputDevice_SDLJoystick(CL_DisplayWindow_SDL *owner, int joystick_num) :
	owner(owner), joystick(0), my_joystick(-1), last_x(0), last_y(0)
{
	type = CL_InputDevice::joystick;

	joystick = SDL_JoystickOpen(joystick_num);
	if (joystick == 0)
	{
		throw CL_Error(SDL_GetError());
	}
	else
	{
		my_joystick = joystick_num;
		slot_sdlevent = owner->sig_sdl_joystick_event.connect(this, &CL_InputDevice_SDLJoystick::handle_joystick_event);
	}
}

CL_InputDevice_SDLJoystick::~CL_InputDevice_SDLJoystick()
{	
	if (joystick != 0)
	{
		SDL_JoystickClose(joystick);
		joystick = 0;
	}
}

/////////////////////////////////////////////////////////////////////////////
// CL_InputDevice_SDLJoystick attributes:

int CL_InputDevice_SDLJoystick::get_x() const
{
	return last_x;
}

int CL_InputDevice_SDLJoystick::get_y() const
{
	return last_y;
}

bool CL_InputDevice_SDLJoystick::get_keycode(int keycode) const
{
	return (SDL_JoystickGetButton(joystick,keycode) == 1);
}

/////////////////////////////////////////////////////////////////////////////
// CL_InputDevice_SDLJoystick operations:

std::string CL_InputDevice_SDLJoystick::get_key_name(int id) const
{
	char buf[256];
	sprintf(buf, "Joystick button %d", id+1);
	return buf;
}

/////////////////////////////////////////////////////////////////////////////
// CL_InputDevice_SDLJoystick implementation:

void CL_InputDevice_SDLJoystick::handle_joystick_axis_move(SDL_Event& event)
{
	CL_InputEvent cl_event;
	cl_event.device = CL_InputDevice(this);
	cl_event.type = CL_InputEvent::moved;
	if(event.jaxis.axis == 0) 
	{
		last_x = event.jaxis.value;
		cl_event.mouse_pos = CL_Point(event.jaxis.value,0);
	}
	else if(event.jaxis.axis == 1) 
	{
		last_y = event.jaxis.value;
		cl_event.mouse_pos = CL_Point(0,event.jaxis.value);
	}
	
	cl_event.device.sig_move().call(cl_event);
}

void CL_InputDevice_SDLJoystick::handle_joystick_hat_move(SDL_Event&)
{
	// todo : Implement hat move
}

void CL_InputDevice_SDLJoystick::handle_joystick_ball_move(SDL_Event&)
{
	// todo : Implement ball move
}

void CL_InputDevice_SDLJoystick::handle_joystick_button(SDL_Event& event)
{
	CL_InputEvent cl_event;
	cl_event.device = CL_InputDevice(this);
	if(event.jbutton.state == SDL_PRESSED) cl_event.type = CL_InputEvent::pressed;
	if(event.jbutton.state == SDL_RELEASED) cl_event.type = CL_InputEvent::released;
	cl_event.id = event.jbutton.button;

	// not sure that this is something the user needs to know
	SDLMod mod = SDL_GetModState();
	cl_event.left_alt = (mod & KMOD_LALT) ? true : false;
	cl_event.right_alt = (mod & KMOD_RALT) ? true : false;
	cl_event.left_shift = (mod & KMOD_LSHIFT) ? true : false;
	cl_event.right_shift = (mod & KMOD_RSHIFT) ? true : false;
	cl_event.left_ctrl = (mod & KMOD_LCTRL) ? true : false;
	cl_event.right_ctrl = (mod & KMOD_RCTRL) ? true : false;

	if (cl_event.type == CL_InputEvent::pressed)
		cl_event.device.sig_key_down().call(cl_event);
	if (cl_event.type == CL_InputEvent::released)
		cl_event.device.sig_key_up().call(cl_event);
}

void CL_InputDevice_SDLJoystick::handle_joystick_event(SDL_Event& event)
{
	switch(event.type)
	{
	case SDL_JOYAXISMOTION:
	{
		if(event.jaxis.which == my_joystick)
			handle_joystick_axis_move(event);
		break;
	}
	case SDL_JOYBUTTONDOWN:
	case SDL_JOYBUTTONUP:
	{	
		if(event.jbutton.which == my_joystick)
			handle_joystick_button(event);
		break;
	}	
	case SDL_JOYBALLMOTION:
	{
		if(event.jball.which == my_joystick)
			handle_joystick_ball_move(event);
		break;
	}
	case SDL_JOYHATMOTION:
	{
		if(event.jhat.which == my_joystick)
			handle_joystick_hat_move(event);
		break;
	}
	}	
}
